import hashlib, base64, tempfile, lib.json, os


class FTP(object):


	def __init__(self, ftp, do):
		self.ftp = ftp
		self.do = do

	def action(self, method, msg, args):
		print(msg % args)
		if self.do : method(*args)

	def chmod(self, mode, path):
		self.action(self.ftp.chmod, "ftp.chmod('%s', '%s')", (mode, path))

	def storbinary(self, path, fd):
		self.action(self.ftp.storbinary, "ftp.storbinary('STOR %s', %s)", (path, fd))

	def mkd(self, path):
		self.action(self.ftp.mkd, "ftp.mkd('%s')", (path,))

	def rmd(self, path):
		self.action(self.ftp.rmd, "ftp.rmd('%s')", (path,))

	def delete(self, path):
		self.action(self.ftp.delete, "ftp.delete('%s')", (path,))

	def rename(self, fr, to):
		self.action(self.ftp.rename, "ftp.rename('%s', '%s')", (fr, to))

	def hascii(self, path):
		hasher = hashlib.sha256()
		self.ftp.retrbinary('RETR %s' % path, hasher.update)
		h = hasher.digest()
		return base64.b64encode(h).decode('ascii')


	def ls( self , path ) :

		for itemtype , item in self.ftp.ls( path ) :

			if item != '.' and item != '..' :

				yield itemtype , item


	def hash(self, root, htree, tree, skip, current = ''):

		for t , item in self.ls( current ) :

			if skip( current , item ) :

				continue

			minipath = current + item

			if t == self.ftp.FILE:
				digest = self.hascii('/%s/%s' % (root, minipath))
				print('%s > %s' % (minipath, digest))

				htree.setdefault(digest, [])
				htree[digest].append(minipath)
				tree[item] = digest

			elif t == self.ftp.DIR:
				tree[item] = {}
				self.hash(root, htree, tree[item], skip, current + item + '/')


	def recursivermd( self , path ) :

		for itemtype , item in self.ls( path ):

			itempath = path + '/' + item

			if itemtype == self.ftp.FILE:
				self.delete( itempath )

			elif itemtype == self.ftp.DIR:
				self.recursivermd( itempath )

		self.rmd( path )


	def _makedirs(self, root, model, current):
		for item, data in model.items():
			if type(data) == dict:
				self.mkd('/%s/%s%s' % (root, current, item))
				self._makedirs(root, model[item], current + item + '/')

	def makedirs(self, root, model, actual, current = ''):

		for item, data in model.items():
			if type(data) == dict:
				if item not in actual:
					self.mkd('/%s/%s%s' % (root, current, item))
					self._makedirs(root, model[item], current + item + '/')
				else:
					self.makedirs(root, model[item], actual[item], current + item + '/')


	def _removedirs(self, root, subtree, current):
		for item, data in subtree.items():
			if type(data) == dict : self._removedirs(root, data, current + '/' + item)

		self.rmd('/%s/%s' % (root, current))


	def removedirs(self, root, model, actual, current = ''):

		for item, entry in actual.items():
			if type(entry) == dict:
				if item not in model:
					self._removedirs(root, entry, current + item)
				else:
					self.removedirs(root, model[item], entry, current + item + '/')


	def sendJSON(self, path, data):
		with tempfile.NamedTemporaryFile('w', delete = False) as tmp:
			lib.json.pretty(data, tmp)

		with open(tmp.name, 'rb') as fd : self.storbinary(path, fd)

		os.remove(tmp.name)




import inspect, lib, os, types

try :
	from importlib import import_module as importmodule
except :
	importmodule = __import__

def reset ( t ) :
	t.__all__ = []

def init ( t ) :
	if type( getattr( t , '__all__' , None ) ) != list :
		reset( t )

def exists(s, t):
	return s in t.__dict__

def public(t, pred = None):
	if pred is None : pred = []

	if len(pred) == 0:
		for key in t.__all__ :
			yield key

	for key in t.__all__:
		for p in pred:
			if p(getattr(t, key)):
				yield key
				break

def setpublic(t, pred = None):
	if pred is None : pred = []

	if len(pred) == 0 :
		t.__all__ = t.__dict__.keys()
		return

	publ = set(t.__all__)
	for key, val in t.__dict__.items():
		if key in publ : continue
		for p in pred :
			if p(val):
				publ.add(key)
				break



	t.__all__ = list(publ)

def ispublic(s, t):
	return s in t.__all__


def setprivate(t, pred = None):
	if pred is None : pred = []

	if len(pred) == 0:
		t.__all__ = []
		return


	publ = set(t.__all__)

	for key in t.__all__:
		for p in pred :
			if p(getattr(t, key)):
				publ.remove(key)
				break

	t.__all__ = list(publ)

def isprivate(s, t):
	return exists(s, t) and not ispublic(s, t)


def clean(t):
	publ = set(t.__all__)
	for key in t.__all__:
		try:
			getattr(t, key)
		except AttributeError:
			publ.remove(key)

	t.__all__ = list(publ)


def package(t):
	reset(t)
	setpublic(t, [inspect.ismodule])

def module(t):
	package(t)
	setpublic(t, [inspect.isclass, inspect.isfunction, inspect.isgenerator])

def toolbox(t):
	reset(t)
	setpublic(t, [inspect.isfunction])


def resolve ( target, module ) :
	return lib.str.mostlikely( target, module.__all__ )


def __init__(t, root, ancestors = None):

	if ancestors is None :
		ancestors = [ ]

	uid = ancestors + [os.path.basename(root)]

	module = '.'.join( uid )

	_all = []

	for f in os.listdir(root):
		path = root + '/' + f

		if os.path.isdir(path):

			if f != '__pycache__' :

				if os.path.isfile( path + '/__init__.py' ) :
					child = importmodule( module + '.' + f )

				else :
					child = types.ModuleType( f )
					__init__( child , path , uid )

				setattr( t , f , child )
				_all.append( f )

		elif os.path.isfile(path) and f != '__init__.py':
			name, ext = os.path.splitext(f)

			if ext == '.py':
				s = importmodule(module + '.' + name)
				setattr(t, name, s)
				_all.append(name)
				toolbox(s)

	t.__all__ = _all

def format ( M , pred ) :
	return ', '.join( o for o , _ in inspect.getmembers( M , pred ) )



import os, lib.file, lib.git, lib.tex

def build(src, out):


	def add(path, f):

		def callback(g):
			lib.file.read(g, f.write)
			f.write(os.linesep)

		if os.path.isdir(path) : lib.file.walk(path, callback)

		elif os.path.isfile('%s.tex' % path):
			with open('%s.tex' % path) as g : callback(g)


	with open(out, 'w') as f:
		a = lambda x : add(os.path.join(src, x), f)
		w = f.write
		a('preamble')
		w('\\begin{document}\n')
		a('title')
		a('before')
		a('main')
		w('\\appendix\n')
		a('appendix')
		a('after')
		w('\\end{document}\n')



def ignore(name):
	for item in lib.tex.out(name):
		print(item)

def clean(name):
	files = lib.tex.out(name)
	lib.file.rm(files)



import lib.sys

BOWER_CMD = "bower"

def runcmd(*args, **kwargs):

	force = kwargs.get( "force", False )

	cmd = []
	cmd.append(BOWER_CMD)
	if force : cmd.append("--force")
	cmd.extend(args)
	lib.sys.call(cmd, stddefault = None)


def register(bowername, gitendpoint, force = False):
	runcmd("register", bowername, "git://%s.git" % gitendpoint, force = force)

def unregister(bowername, force = False):
	runcmd("unregister", bowername, force = force)



import lib.ghbadges as ghb

TRAVISCI = "travis-ci"
DRONEIO = "drone.io"

CI = [TRAVISCI, DRONEIO]

def qualifiedname ( username, repo ) :
	return "%s-%s" % ( username, repo )


def formatargs ( username, repo ) :

	return dict(
		username = username,
		repo = repo,
		packagename = qualifiedname( username, repo )
	)


DEFAULT = ghb.DEFAULT
FLAT = ghb.FLAT
FLATSQUARED = ghb.FLATSQUARED

JPG = ghb.JPG
PNG = ghb.PNG
SVG = ghb.SVG


def badges ( username, repo, ci, cb, style = FLAT, fmt = SVG ) :

	fmtargs = formatargs( username, repo )

	vendors = [
		ghb.npmlicense,
		ghb.npmversion,
		ghb.bowerversion,
		ghb.traviscibuild if ci == TRAVISCI else ghb.droneiobuild,
		ghb.coverallscoverage,
		ghb.daviddependencies,
		ghb.daviddevdependencies,
		ghb.codeclimategpa,
		ghb.npmdownloads,
		ghb.githubissues
	]

	for vendor in vendors :
		cb( vendor( style = style, fmt = fmt, **fmtargs ) )

def installationinstructions ( username , name ) :

	"""

		>>> import lib.codebricks
		>>> print( lib.codebricks.installationinstructions( "abcd" , "efgh" ) )
		Can be managed through [duo](https://github.com/duojs/duo),
		[component](https://github.com/componentjs/component),
		[bower](https://github.com/bower/bower), or
		[npm](https://github.com/npm/npm).
		<BLANKLINE>
		```js
		let efgh = require( "abcd-js-efgh" ) ;
		```

	"""

	return r"""Can be managed through [duo](https://github.com/duojs/duo),
[component](https://github.com/componentjs/component),
[bower](https://github.com/bower/bower), or
[npm](https://github.com/npm/npm).

```js
let {1} = require( "{0}-js-{1}" ) ;
```""".format( username , name )



import lib.json , itertools , lib.error , inspect

ARGS = "args"
KWARGS = "kwargs"
NO = "no"
DIRECTIVE_ARG = "+a"

def splitparts ( string , argv , i = 0 , j = None , eos = None , buf = "" , escapefirst = False ) :

	"""
		Splits a command line argument string into tokens. Tokens are separated
		by whitespace. Whitespace can be included inside tokens by wrapping tokens with
		single or double quotes. Whitespace can also be included by escaping them
		without the need for quote wrapping. When inside a string standard escape
		sequence are allowed : line-feed and tabulations for example. This method will
		return a tuple that can be used to continue the analysis of a string that was
		incomplete the first time the method was run.
	"""

	if j == None :
		j = len( string )

	if escapefirst and i < j :
		buf += string[i]
		i += 1

	while i < j :

		c = string[i]

		if c == '\\' :
			# escaped character
			i += 1

			if i == j :
				return eos , buf , True

			buf += string[i]

		elif eos is None and c == ' ' :
			if buf :
				argv.append( buf )
				buf = ""

		elif c == '\'' or c == '"' :
			if eos is None :
				# beginning of a new string
				eos = c
			elif c == eos :
				# this is the end of the current string
				eos = None
			else :
				# this is simply an quote character
				buf += c

		else :
			buf += c

		i += 1

	return eos , buf , False


def split ( string ) :

	argv = []

	eos , buf , escaped = splitparts( string , argv )

	if eos is not None:
		raise Exception( "could not split arguments : incomplete string" )

	elif escaped :
		raise Exception( "could not split arguments : trailing escape char" )

	if buf :
		argv.append( buf )

	return argv


def parse ( argv, args, kwargs ) :

	"""

		>>> from lib.args import *

		>>> parse( ["a", "b", "c", "--flag"], [], {} )
		(['a', 'b', 'c'], {'flag': True})

	"""

	key = ""
	isflag = False
	isvalue = False
	islist = False
	isarg = False

	for p in argv:

		if isarg : args.append(p)

		elif p == DIRECTIVE_ARG : isarg = True

		elif len(p) > 1 and p[0] == '-' and not p[1].isdigit():
			isvalue = False
			isflag = False
			islist = False
			if p[1] == '-':
				p = p[2:]

				if len(p) == 0 : continue

				v = True
				if len(p) > 1 and p[:2] == NO :
					v = False
					p = p[2:]

				kwargs[p] = v
				key = p
				isflag = True

			elif len(p) == 2:

				p = p[1]

				kwargs[p] = True
				key = p
				isflag = True

			else:
				for c in p[1:] : kwargs[c] = True

		else :
			if isflag :
				isflag = False
				isvalue = True
				kwargs[key] = p
			elif isvalue :
				isvalue = False
				islist = True
				kwargs[key] = [kwargs[key], p]
			elif islist : kwargs[key].append(p);
			else : args.append(p)

	return args, kwargs


def read ( string , args , kwargs ) :

	argv = split( string )

	return parse( argv , args , kwargs )


def format ( key, val ) :

	"""

		>>> from lib.args import *

		>>> format( 'filter', 'false' )
		'--filter=false'

		>>> format( 'i', 'input.json' )
		'-i=input.json'

	"""

	fmt = "-%s=%s"

	if len(key) > 1 : fmt = "-" + fmt

	return fmt % (key, val)

def escape ( val , quotes = "\"" ) :

	"""

		>>> from lib.args import *

		>>> escape( 'abcd' )
		'"abcd"'

		>>> escape( 'abcd"' )
		'"abcd\\""'

		>>> escape( 'abcd\'' , quotes = "'" )
		"'abcd\\''"

		>>> escape( 'ab\\cd\'' , quotes = "'" )
		"'ab\\\\cd\\''"

	"""

	out = ""

	for c in val :

		if c == "\\" or c == quotes : out += "\\"

		out += c

	return quotes + out + quotes

def listify ( arg ) :

	"""

		>>> from lib.args import *

		>>> listify( None )
		[]

		>>> listify( 'value' )
		['value']

		>>> listify( ['value1', 'value2'] )
		['value1', 'value2']

	"""

	if arg is None :
		return []
	elif isinstance( arg, str ) :
		return [ arg ]
	else :
		return arg


def kwargslist( spec ) :

	out = []

	if spec.args is not None and spec.defaults is not None :
		out.extend( spec.args[-len( spec.defaults ):] )

	if spec.kwonlyargs :
		out.extend( spec.kwonlyargs )

	if spec.varargs is not None :
		out.append( spec.varargs )

	if spec.varkw is not None :
		out.append( spec.varkw )

	return out



def inflate ( args, kwargs ) :
	"""
		Loads more arguments if JSON ARGS or KWARGS source have been specified.
		Since JSON KWARGS could contain more ARGS or KWARGS directives
		the inflate operations loops until no JSON source is specified anymore.
		The last JSON source that is read will always serve as a
		base and thus JSON files referencing other JSON files are inheriting the
		properties of those other JSON files. Similarly, JSON ARGS will be prepended to the
		existing argument list.

		/!\ Currently only support one kwargs json file argument
	"""

	jsonargssources = []

	while True :

		if ARGS in kwargs :
			if not isinstance(kwargs[ARGS], list) : kwargs[ARGS] = [kwargs[ARGS]]
			jsonargssources[:0] = kwargs[ARGS]
			del kwargs[ARGS]

		if not KWARGS in kwargs : break

		source = kwargs[KWARGS]
		del kwargs[KWARGS]

		kwargscopy = dict(kwargs)
		kwargs.clear()

		with lib.json.proxy(source, throws = True) as data : kwargs.update(data)

		kwargs.update(kwargscopy)


	argscopy = list(args)
	del args[:]

	for source in jsonargssources :
		with lib.json.proxy(source, throws = True) as data : args.extend(data)

	args.extend(argscopy)


def pairs ( fn , args , kwargs ) :

	argkeys = fn.__code__.co_varnames

	argpairs = ( ( argkeys[i], v ) for i, v in enumerate( args ) )
	kwargpairs = kwargs.items()

	return itertools.chain( argpairs, kwargpairs )


def names ( fn , args , kwargs ) :

	return set( [ name for name , _ in lib.args.pairs( fn , args , kwargs ) ] )


def mandatory ( handle = lib.error.throws(Exception), **names ) :

	"""

		>>> from lib.args import *

		>>> @mandatory( print, c = True, b = False )
		... def test ( a = 12, b = None, c = None ) :
		... 	print( 'ok' )

		>>> test ( 13, c = 'abdf', b = [1] )
		ok

		>>> test ( 13.2, b = [1,2] )
		missing mandatory argument 'c'
		ok

		>>> test ( 13, c = 'df' )
		ok

		>>> test ( b = 2 , c = 'df' )
		ok

	"""

	def wrap ( fn ) :

		def wrapper( *args, **kwargs ) :

			fmt = "missing mandatory argument '%s'"

			keys = lib.args.names( fn , args , kwargs )

			for k, v in names.items() :

				if v and not k in keys :

					handle( fmt % k )

			return fn( *args, **kwargs )

		wrapper.__name__ = fn.__name__
		return wrapper

	return wrap


def validate ( handle = lib.error.throws(Exception), **predicates ) :

	"""

		>>> from lib.args import *

		>>> @validate( print, a = lambda v : v <= 13 , b = lambda v : len(v) > 0, c = lambda v : v[:2] == "ab" )
		... def test ( a, b = None, c = None ) :
		... 	print( 'ok' )

		>>> test ( 13, c = 'abdf', b = [1] )
		ok

		>>> test ( 13.2, c = 'abdf', b = [1,2] )
		cannot validate arg 'a' = 13.2
		ok

		>>> test ( 13, c = 'df' )
		cannot validate arg 'c' = 'df'
		ok

	"""

	def wrap ( fn ) :

		def wrapper( *args, **kwargs ) :

			fmt = "cannot validate arg '%s' = %r"

			for k, v in lib.args.pairs( fn , args , kwargs ) :

				if k in predicates and not predicates[k](v) :
					handle( fmt % ( k , v ) )

			return fn( *args, **kwargs )

		wrapper.__name__ = fn.__name__
		return wrapper

	return wrap

def accepts ( handle = lib.error.throws(Exception), **types ) :

	"""

		>>> from lib.args import *

		>>> @accepts( print, a = int, b = list, c = str )
		... def test ( a, b = None, c = None ) :
		... 	print( 'ok' )

		>>> test ( 13, c = 'df', b = [] )
		ok

		>>> test ( 13.2, c = 'df', b = [] )
		arg 'a' = 13.2 does not match <class 'int'>
		ok

		>>> test ( 13, c = 'df' )
		ok

	"""

	def wrap ( fn ) :

		def wrapper( *args, **kwargs ) :

			fmt = "arg '%s' = %r does not match %s"

			for k, v in lib.args.pairs( fn , args , kwargs ) :

				if k in types and not isinstance( v, types[k] ) :
					handle( fmt % ( k, v, types[k] ) )

			return fn( *args, **kwargs )

		wrapper.__name__ = fn.__name__
		return wrapper

	return wrap


def convert ( handle = lib.error.throws( Exception ), **convertors ) :

	"""

		>>> from lib.args import *

		>>> @convert( print, a = int, b = list, c = str )
		... def test ( a, b = list("abc"), c = "2" ) :
		... 	print( [a , b , c] )

		>>> test ( "13", c = 5, b = "123" )
		[13, ['1', '2', '3'], '5']

		>>> test ( 13.2, c = 'df', b = (1,) )
		[13, [1], 'df']

		>>> test ( 13, c = 'df' )
		[13, ['a', 'b', 'c'], 'df']

		>>> test ( 13, 1, c = 'df' )
		error in the conversion of arg 'b' = 1 using <class 'list'>, the error is 'int' object is not iterable
		[13, 1, 'df']

	"""

	def wrap ( fn ) :

		def wrapper( *args, **kwargs ) :

			argkeys = fn.__code__.co_varnames

			argpairs = ( ( i, argkeys[i], v ) for i, v in enumerate( args ) )
			kwargpairs = ( ( k , k , v ) for k , v in kwargs.items() )
			m = len ( args )
			n = len ( kwargs )

			fmt = "error in the conversion of arg '%s' = %r using %s, the error is %s"

			args = list( args )
			kwargs = dict( kwargs )

			for a, ikv in zip( itertools.chain( itertools.repeat(args, m ), itertools.repeat(kwargs, n) ),  itertools.chain( argpairs, kwargpairs ) ) :

				i, k, v = ikv

				if k in convertors :
					try :
						 a[i] = convertors[k]( v )
					except Exception as e:
						handle( fmt % ( k, v, convertors[k], e ) )

			return fn( *args, **kwargs )

		wrapper.__name__ = fn.__name__
		return wrapper

	return wrap


def forward ( callable , locals ) :

	argspec = inspect.getfullargspec( callable )

	if argspec.args is None : args = []
	else : args = [ locals[key] for key in argspec.args ]

	if argspec.varkw is None : kwargs = {}
	else : kwargs = { key : item for key , item in locals.items( ) if key in argspec.varkw }

	return callable( *args , **kwargs )








def image ( title, img ) :
	return "![%s](%s)" % ( title, img )


def link ( text, href ) :
	return "[%s](%s)" % ( text, href )


def imagewithlink ( title, img, href ) :
	text = image( title, img )
	return link( text, href )




import json, os.path, inspect, lib.error, lib.kwargs, functools


pretty = functools.partial(json.dump, indent = '\t', separators = (',', ': '))

class proxy(object):

	def __init__(self, fname, mode = 'r', default = None, throws = False, **kwargs):
		self.fname = fname
		self.data  = default if mode == 'r' or default is not None else {}
		self.mode  = mode
		self.throws = throws
		self.kwargs = kwargs
		self.bytes = ""

	def backup(self):
		with open(self.fname, 'r') as f : self.bytes = f.read()

	def restore(self):
		with open(self.fname, 'w') as f : f.write(self.bytes)

	def load(self):
		kwargs = lib.kwargs.filter(self.kwargs, json.JSONDecoder)
		with open(self.fname, 'r') as f : self.data = json.load(f, **kwargs)

	def dump(self):
		kwargs = lib.kwargs.filter(self.kwargs, json.dump)
		with open(self.fname, 'w') as f : pretty(self.data, f, **kwargs)

	def __enter__(self):
		if os.path.exists(self.fname) :
			self.load()
			self.backup()
		elif self.throws :
			raise lib.error.FileDoesNotExist(self.fname)
		return self.data

	def __exit__(self, t, value, traceback):
		if self.mode == 'w':
			try : self.dump()
			except Exception as e :
				self.restore()
				raise e




import os, bisect


def split(item):
	x = 0

	for i, c in enumerate(item):

		if not c.isdigit() : break

		x *= 10
		x += int(c)

	return x, i


def read(cwd):
	d = {}
	w = 0
	for f in os.listdir(cwd):
		x, i = split(f)
		l = d.setdefault(x, [])
		v = f[i:]
		l.append((f, v))
		w = max(w, i)

	return d, w


def format(x, z):
	return ''.join(['{0:0', z, 'd}']).format(x)


def shift(d, n, beg = 0, end = None):

	out = []
	w = 0

	if len(d) == 0 or n == 0 : return out, w

	keys = sorted(d.keys())

	length = len(keys)

	if end is None : end = keys[-1] + 1

	i = bisect.bisect_left(keys, beg)

	while i < length:

		k = keys[i]
		if k >= end : break

		_k = str(k + n)
		l = d[k]
		out += [(f, _k + v) for f, v in l]

		w = max(w, len(_k))

		i += 1


	return out, w


def drange(where, beg, end):

	where = int(where)
	beg = int(beg)

	if end is not None : end = int(end)

	return where, beg, end





import lib.queue

def stat ( ) :

	for key , val in lib.queue.stat( ) :

		print( key )


def submit ( name , cmd , nodes = None , cpu = None , output = None , error = None , walltime = None , memory = None , redirect = False ) :

	out , err , p = lib.queue.submit( name , cmd , nodes , cpu , output , error , walltime , memory , redirect )

	if out : print( out.decode() , end = "" )
	if err : print( err.decode() , end = "" )



"""
	git module

	some git snippets freely derived from https://github.com/badele/gitcheck
"""


import os, subprocess, re, functools, lib.sys


def info(*args):

	def callback(d):
		lo = helper.locals(d)
		fr = {}
		to = {}

		branch = helper.branch(d)

		for remote in helper.remotes(d):
			if helper.hasbranch(d, remote, branch):
				fr[remote] = helper.pulls(d, remote, branch)
				to[remote] = helper.pushs(d, remote, branch)

		helper.pretty(d, lo, fr, to)

	return helper.wrap(args, callback)


def do(action, *args):

	def callback(d):
		print('%s \'%s\'' % (action, d))
		getattr(helper, action)(d)

	return helper.wrap(args, callback)

def update(*args):
	return do('update', *args)

def up(*args):
	return do('up', *args)

def pull(*args):
	return do('pull', *args)


def commit(*cmplxargs):

	args, dirs = helper.parsecmplx(*cmplxargs)

	def callback(d):
		print('%s \'%s\'' % ("commit", d))
		helper.commit(d, *args)

	return helper.wrap(dirs, callback)

def add(*cmplxargs):

	args, dirs = helper.parsecmplx(*cmplxargs)

	def callback(d):
		print('%s \'%s\'' % ("add", d))
		helper.add(d, *args)

	return helper.wrap(dirs, callback)


def push(*args):
	return do('push', *args)

def ls(*args):
	return do('ls', *args)

def count(*args):

	n = [0]

	def callback(d):
		n[0] += 1

	helper.wrap(args, callback)

	print(n[0])





class helper(object):

	class exception(Exception):
		def __init__(self, what):
			Exception.__init__(self, what)

		def what(self):
			return self.args[0]

		def __repr__(self):
			return self.what()

	def check(args):
		for d in args:
			if not os.path.isdir(d):
				raise helper.exception('[Errno 2] No such file or directory: \'%s\'' % d)

	def wrap(args, callback):

		try:
			helper.check(args)
			helper.recursive(args, callback)

		except helper.exception as e:
			print(e)

	def parsecmplx(*cmplxargs):

		try :
			i = list(cmplxargs).index("--")
		except Exception as e:
			i = len(cmplxargs)

		return cmplxargs[:i], cmplxargs[i+1:]


	def recursive(args, callback):
		if len(args) == 0 : args = ['.']

		for d in args :

			try :

				l = [x for x in os.listdir(d) if os.path.isdir(os.path.join(d, x))]
				if len(l) == 0 : continue
				if '.git' in l : callback(d)
				else           : helper.recursive([os.path.join(d, x) for x in l], callback)

			except Exception as e :

				print( "skipping" , d , "because" , e )


	BRANCH = re.compile(r'^\* (.*)')

	def branch(d):
		out, _, _ = lib.sys.call(['git', 'branch'], cwd = d)
		m = helper.BRANCH.match(out.decode())
		return None if m is None else m.group(1)

	def remotes(d):
		out, _, _ = lib.sys.call(['git', 'remote'], cwd = d)
		return out.decode().split('\n')[:-1]

	def update(d):
		subprocess.call(['git', 'remote', 'update'], cwd = d)

	def up(d):
		subprocess.call(['git', 'up'], cwd = d)

	def commit(d, *args):
		subprocess.call(['git', 'commit'] + list(args), cwd = d)

	def add(d, *args):
		subprocess.call(['git', 'add'] + list(args), cwd = d)

	def push(d):
		subprocess.call(['git', 'push'], cwd = d)

	def pull(d):
		subprocess.call(['git', 'pull'], cwd = d)

	def ls(d):
		pass

	def pulls(d, remote, branch):
		return helper.commits(d, 'HEAD', '%s/%s' % (remote, branch))

	def pushs(d, remote, branch):
		return helper.commits(d, '%s/%s' % (remote, branch), 'HEAD')

	def commits(d, a, b):
		out, _, _ = lib.sys.call(['git', 'log', '%s..%s' % (a, b),  '--oneline'], cwd = d)
		return out.decode().split('\n')[:-1]

	def locals(d):
		out, _, _ = lib.sys.call(['git', 'status', '--porcelain', '-u'], cwd = d)
		return [c for c in out.decode().split('\n')[:-1]]

	def hasbranch(d, remote, branch):
		out, _, _ = lib.sys.call(['git', 'branch', '-r'], cwd = d)
		return out.decode().find('%s/%s' % (remote, branch)) != -1


	def pretty(d, lo, fr, to):
		r = [] if len(lo) == 0 else ['local[%d]' % len(lo)]
		for key in fr :
			if len(fr[key]) > 0 or len(to[key]) > 0:
				r.append('%s[pull:%d, push:%d]' % (key, len(fr[key]), len(to[key])))

		if len(r) > 0 : print(' '.join([d] + r))




__all__ = []

def __init__():
	import os
	root = os.path.dirname(os.path.abspath(__file__))
	module = os.path.basename(root)

	for f in os.listdir(root):
		path = root + '/' + f

		if os.path.isdir(path):
			if os.path.isfile(path + '/__init__.py'):
				__import__('lib.' + module + '.' + f)
				__all__.append(f)

		elif os.path.isfile(path) and f != '__init__.py':
			name, ext = os.path.splitext(f)

			if ext == '.py':
				__import__('lib.' + module + '.' + name)
				__all__.append(name)

__init__()



import lib.error, lib.url

SPOOF_USER_AGENT = "Mozilla/5.0 (X11; U; Linux x86_64; fr; rv:1.9.1.5) Gecko/20091109 Ubuntu/9.10 (karmic) Firefox/3.5.5"

try:

	import lxml.etree

	try :
		import urllib.request as urllib2
	except :
		import urllib2


	def title ( url, *args ) :
		request = urllib2.Request(url)
		request.add_header("User-Agent", SPOOF_USER_AGENT)
		opener = urllib2.build_opener(urllib2.HTTPCookieProcessor)
		conn = opener.open(request)
		parser = lxml.etree.HTMLParser(encoding = "utf-8")
		tree = lxml.etree.parse(conn, parser = parser)
		title = tree.find('.//title')
		if title is not None : text = title.text
		else : text = url

		print ( text )

		if len(args) : href(*args)


	def href(url, *args):
		request = urllib2.Request(url)
		request.add_header("User-Agent", SPOOF_USER_AGENT)
		opener = urllib2.build_opener(urllib2.HTTPCookieProcessor)
		conn = opener.open(request)
		parser = lxml.etree.HTMLParser(encoding = "utf-8")
		tree = lxml.etree.parse(conn, parser = parser)
		title = tree.find('.//title')
		if title is not None : text = title.text
		else : text = url
		fmt = '<a href="%s">%s</a>'
		print(fmt % (url, text))

		if len(args) : href(*args)

except ImportError as cause:

	e = lib.error.ModuleMissingException(cause, "lxml")

	title = lambda url, *args : lib.error.throw(e)
	href = lambda url, *args : lib.error.throw(e)


def asgetparams ( **kwargs ) :
	print( lib.url.get( **kwargs ) )




import os.path , inspect , lib.check , lib.pacman , lib.args , lib.fn , lib.str

ROOT = os.path.normpath(os.path.join(os.path.dirname(__file__), ".."))
BIN = os.path.join(ROOT, "bin")
LIB = os.path.join(ROOT, "lib")
SAK = os.path.join(ROOT, "sak")
DATA = os.path.join(ROOT, "data")

def data(*path):
	return os.path.join(DATA, *path)

def findaction ( module , inp , hierarchy ) :

	# DETERMINE MODULE

	i = 0

	while True :

		parent = '.'.join(hierarchy)

		lib.check.ModuleOrActionNameSpecified( parent , module , inp , i )
		moduleName = inp[i]

		modules = lib.pacman.resolve( moduleName , module )

		lib.check.ModuleOrActionNameExists( parent, module , moduleName , modules )
		lib.check.ModuleOrActionNameNotAmbiguous( parent, moduleName, modules )

		moduleName = modules[0]
		module = getattr( module , moduleName )

		hierarchy.append( moduleName )

		i += 1

		if lib.fn.callable( module ) : break

	return hierarchy , module , inp[i:]


def resolve ( module , inp , hierarchy ) :

	# DETERMINE MODULE

	for part in inp :

		parent = '.'.join(hierarchy)

		modules = lib.pacman.resolve( part , module )

		lib.check.ModuleOrActionNameExists( parent, module , part , modules )
		lib.check.ModuleOrActionNameNotAmbiguous( parent, part, modules )

		part = modules[0]
		module = getattr( module , part )

		hierarchy.append( part )

	return hierarchy , module


def assignarguments ( hierarchy, action, inp ) :

	# CHECK ACTION ARGUMENTS

	moduleName = ".".join( hierarchy[:-1] )
	actionName = hierarchy[-1]

	args, kwargs = lib.args.parse( inp, [], {} )

	spec = inspect.getfullargspec( action )

	kwargslist = lib.args.kwargslist( spec )

	if kwargs :

		lib.check.KwargsNotSupportedException( actionName, kwargslist )

		_kwargs = dict()

		for kwarg in kwargs :

			matching = lib.str.mostlikely( kwarg, kwargslist )

			if not matching and spec.varkw :
				_kwargs[kwarg] = kwargs[kwarg]
			else :
				lib.check.KwargNameExists( kwarg, actionName, matching, kwargslist )
				lib.check.KwargNameNotAmbiguous( kwarg, actionName, matching )

				_kwargs[matching[0]] = kwargs[kwarg]

		kwargs = _kwargs

	# WE INFLATE AFTER RESOLVING KWARGS THUS
	# KWARGS IN JSON ARGS FILES MUST BE
	# EXACTLY MATCHING SPECS OF ACTIONS

	lib.args.inflate( args, kwargs )

	m = ( 0 if spec[0] is None else len( spec[0] ) ) -\
	    ( 0 if spec[3] is None else len( spec[3] ) )
	n = len( args )

	lib.check.NotTooFewArgumentsForAction( moduleName, actionName, n, m, spec )
	lib.check.NotTooManyArgumentsForAction( moduleName, actionName, n, m, spec )


	# DONE
	return args, kwargs





"""
	Exposes Github API v3.
"""


import json , functools
import lib.fn , lib.args
import lib.config, lib.git, lib.error, lib.check, lib.curl, lib.url , lib.dict

# CONSTANTS

DOMAIN = 'github.com'
CONFIG_KEY = 'github'

YOU = "you"
USER = "user"
ORG = "org"

TARGET = "target"
TARGETS = [YOU, USER, ORG]

ALL = "all"
OWNER = "owner"
PUBLIC = "public"
PRIVATE = "private"
MEMBER = "member"
FORKS = "forks"
SOURCES = "sources"

TYPE = "t"
TYPES_YOU = [ALL, OWNER, PUBLIC, PRIVATE, MEMBER]
TYPES_USER = [ALL, OWNER, MEMBER]
TYPES_ORG  = [ALL, PUBLIC, PRIVATE, FORKS, SOURCES, MEMBER]

TYPES = {
	YOU : TYPES_YOU,
	USER : TYPES_USER,
	ORG  : TYPES_ORG
}

TYPES_DEFAULT = {
	YOU : ALL,
	USER : OWNER,
	ORG  : ALL
}



# TOOLS

def api ( *args ) :
	return "https://api.github.com/" + '/'.join( map( str , args ) )

@lib.fn.throttle( 20 , 70 )
def send ( method , url , data = None , **kwargs ) :

	"""
		Throttling because
		https://github.com/octokit/octokit.net/issues/638#issuecomment-67795998
	"""

	contenttype = "application/json"

	if data is not None :
		data = json.dumps(data)

	return lib.curl.call( method , api( *url ) , contenttype , data = data , **kwargs )

put = functools.partial( send , lib.curl.PUT )
get = functools.partial( send , lib.curl.GET )
post = functools.partial( send , lib.curl.POST )
update = functools.partial( send , lib.curl.UPDATE )
patch = functools.partial( send , lib.curl.PATCH )
delete = functools.partial( send , lib.curl.DELETE )


def credentials ( username = None , password = None ) :

	return lib.config.prompt_cred( DOMAIN , CONFIG_KEY , username , password )


def paginate ( url , username = None , password = None ) :

	username , password = credentials( username , password )

	pageid = 1

	while True :

		pageurl = url[:-1] + ( url[-1] + lib.url.get( page = str( pageid ) ) , )

		out , err , p = get( pageurl , username = username , password = password )

		lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )

		pagecontent = json.loads( out.decode() )

		if not pagecontent : break

		yield pagecontent

		pageid += 1


def itemize ( url , username = None , password = None ) :

	for page in paginate( url , username = username , password = password ) :

		for item in page : yield item


def validate ( data ) :

	if "message" in data :

		msg = data["message"]

		if "errors" in data :

			for i , err in enumerate( data["errors"] ) :

				msg += "\n#" + str( i ) + " -> " + json.dumps( err )

		raise lib.error.GithubAPIException( msg )



# REPOS

def list ( target = YOU, name = None, t = None, username = None, password = None ) :

	lib.check.OptionNotInListException( TARGET, target, TARGETS )

	if t is None : t = TYPES_DEFAULT[target]
	lib.check.OptionNotInListException( TYPE, t, TYPES[target] )

	if target == YOU or t == PRIVATE or username is not None :
		username , password = credentials( username , password )

	if   target ==  YOU : url = ( "user" , "repos" )
	elif target == USER : url = ( "users" , name , "repos" )
	elif target ==  ORG : url = ( "orgs" , name , "repos" )

	return itemize( url , username = username , password = password )



# ISSUES

def issues ( owner = None , repo = None , number = None , user = False , org = None , username = None , password = None , milestone = None , filter = None , state = None , creator = None , assignee = None , mentioned = None , labels = None , sort = None , direction = None , since = None ) :

	"""
		https://developer.github.com/v3/issues/
	"""

	username , password = credentials( username , password )

	if owner is not None and repo is not None :
		if number is None :
			url = ( "repos" , owner , repo , "issues" )
		else :
			url = ( "repos" , owner , repo , "issues" , number )
	elif user :
		url = ( "user" , "issues" )
	elif org is not None :
		url = ( "orgs" , org , "issues" )
	else :
		url = ( "issues" , )

	keys = [ "milestone" , "filter" , "state" , "assignee" , "creator" , "mentioned" , "labels" , "sort" , "direction" , "since" ]

	parameters = lib.dict.select( locals( ) , keys )

	out , err , p = get( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )

	return json.loads( out.decode( ) )


def closeissues ( owner , repo , *issuenos , username = None , password = None ) :

	username , password = credentials( username , password )

	for number in issuenos :

		issue = issues( owner , repo , number , username = username , password = password )

		keys = [ "title" , "body" , "assignee" , "milestone" , "labels" ]

		parameters = lib.dict.select( issue , keys )

		parameters["state"] = "closed"

		yield editissue( owner , repo , number , username = username , password = password , **parameters )


def createissue ( owner , repo , title , body = None , assignee = None , milestone = None , labels = None , username = None , password = None ) :

	"""
		https://developer.github.com/v3/issues/#create-an-issue
	"""

	url = ( "repos" , owner , repo , "issues" )

	labels = lib.args.listify( labels )

	keys = [ "title" , "body" , "assignee" , "milestone" , "labels" ]

	parameters = lib.dict.select( locals( ) , keys )

	username , password = credentials( username , password )

	out , err , p = post( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )

	return json.loads( out.decode( ) )


def editissue ( owner , repo , number , title = None , body = None , assignee = None , state = None , milestone = None , labels = None , username = None , password = None ) :

	"""
		https://developer.github.com/v3/issues/#edit-an-issue
	"""

	url = ( "repos" , owner , repo , "issues" , number )

	labels = lib.args.listify( labels )

	keys = [ "title" , "body" , "assignee" , "state" , "milestone" , "labels" ]

	parameters = lib.dict.select( locals( ) , keys )

	username , password = credentials( username , password )

	out , err , p = patch( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )



# LABELS

def labels ( owner, repo, name = None, issue = None, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	if issue is not None :
		url = ( "repos", owner, repo, "issues", issue, "labels" )
	elif name is not None :
		url = ( "repos", owner, repo, "labels", name )
	else :
		url = ( "repos", owner, repo, "labels" )

	out , err , p = get( url, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )


def createlabel ( owner, repo, name, color, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	url = ( "repos", owner, repo, "labels" )

	parameters = lib.dict.select( locals( ) , [ "name" , "color" ] )

	username , password = credentials( username , password )

	out , err , p = post(url, data = parameters, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )


def updatelabel ( owner, repo, oldname, newname, color, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	url = ( "repos", owner, repo, "labels", oldname )

	parameters = dict( name = newname, color = color )

	username, password = credentials( username, password )

	out , err , p = patch( url, data = parameters, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )



def deletelabel ( owner, repo, name, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	url = ( "repos", owner, repo, "labels", name )

	username, password = credentials( username, password )

	out , err , p = delete( url, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )


def addlabels ( owner, repo, issue, labels = None, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	labels = lib.args.listify( labels )

	username, password = credentials( username, password )

	url = ( "repos", owner, repo, "issues", issue, "labels" )

	out , err , p = post( url, data = labels, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )


def removelabel ( owner, repo, issue, label, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	username, password = credentials( username, password )

	url = ( "repos", owner, repo, "issues", issue, "labels", label )

	out , err , p = delete( url, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )


def updatelabels ( owner, repo, issue, labels = None, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	labels = lib.args.listify( labels )

	username, password = credentials( username, password )

	url = ( "repos", owner, repo, "issues", issue, "labels" )

	out , err , p = put( url, data = labels, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )


def removealllabels ( owner, repo, issue, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	username, password = credentials( username, password )

	url = ( "repos", owner, repo, "issues", issue, "labels" )

	out , err , p = delete( url, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )


def milestonelabels ( owner, repo, milestone, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	url = ( "repos", owner, repo, "milestones", milestone, "labels" )

	out , err , p = get( url, username = username, password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )
	return json.loads( out.decode( ) )



# MILESTONES

def milestones ( owner , repo , number = None , state = None , sort = None , direction = None , username = None , password = None ) :

	if number is None :
		url = ( "repos" , owner , repo , "milestones" )
	else :
		url = ( "repos" , owner , repo , "milestones" , number )

	parameters = lib.dict.select( locals( ) , [ "state" , "sort" , "direction" ] )

	username , password = credentials( username , password )

	out , err , p = get( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args , p.returncode )
	return json.loads( out.decode( ) )


def createmilestone ( owner , repo , title , state = None , description = None , due_on = None , username = None , password = None ) :

	url = ( "repos" , owner , repo , "milestones" )

	parameters = lib.dict.select( locals( ) , [ "title" , "state" , "description" , "due_on" ] )

	username , password = credentials( username , password )

	out , err , p = post( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args , p.returncode )
	return json.loads( out.decode( ) )


def updatemilestone ( owner , repo , number , title , state = None , description = None , due_on = None , username = None , password = None ) :

	url = ( "repos" , owner , repo , "milestones" , number )

	parameters = lib.dict.select( locals( ) , [ "title" , "state" , "description" , "due_on" ] )

	username , password = credentials( username , password )

	out , err , p = patch( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args , p.returncode )
	return json.loads( out.decode( ) )


def deletemilestone ( owner , repo , number , username = None , password = None ) :

	url = ( "repos" , owner , repo , "milestones" , number )

	username , password = credentials( username , password )

	out , err , p = delete( url , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args , p.returncode )
	return json.loads( out.decode( ) )



# COMMENTS

def comments ( owner , repo , id = None , number = None , sort = None , direction = None , since = None , username = None , password = None ) :

	if id is not None :
		url = ( "repos" , owner , repo , "issues" , "comments" , id )
	elif number is None :
		url = ( "repos" , owner , repo , "issues" , "comments" )
	else :
		url = ( "repos" , owner , repo , "issues" , number , "comments" )

	parameters = lib.dict.select( locals( ) , [ "sort" , "direction" , "since" ] )

	username , password = credentials( username , password )

	out , err , p = get( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args , p.returncode )
	return json.loads( out.decode( ) )


def createcomment ( owner , repo , number , body , username = None , password = None ) :

	url = ( "repos" , owner , repo , "issues" , number , "comments" )

	parameters = lib.dict.select( locals( ) , [ "body" ] )

	username , password = credentials( username , password )

	out , err , p = post( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args , p.returncode )
	return json.loads( out.decode( ) )


def editcomment ( owner , repo , id , body , username = None , password = None ) :

	url = ( "repos" , owner , repo , "issues" , "comments" , id )

	parameters = lib.dict.select( locals( ) , [ "body" ] )

	username , password = credentials( username , password )

	out , err , p = patch( url , data = parameters , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args , p.returncode )
	return json.loads( out.decode( ) )


def deletecomment ( owner , repo , id , username = None , password = None ) :

	url = ( "repos" , owner , repo , "issues" , "comments" , id )

	username , password = credentials( username , password )

	out , err , p = delete( url , username = username , password = password )
	lib.check.SubprocessReturnedFalsyValueException( p.args , p.returncode )
	return json.loads( out.decode( ) )





LICENSES = [
	None,
	"agpl-3.0",
	"apache-2.0",
	"artistic-2.0",
	"bsd-2-clause",
	"bsd-3-clause",
	"cc0",
	"epl-1.0",
	"gpl-2.0",
	"gpl-3.0",
	"isc",
	"lgpl-2.1",
	"lgpl-3.0",
	"mit",
	"mpl-2.0",
	"no-license",
	"unlicense"
]

GITIGNORES = [
	None,
	"Actionscript",
	"Ada",
	"Agda",
	"Android",
	"AppceleratorTitanium",
	"ArchLinuxPackages",
	"Autotools",
	"Bancha",
	"CakePHP",
	"CFWheels",
	"C",
	"C++",
	"ChefCookbook",
	"Clojure",
	"CMake",
	"CodeIgniter",
	"CommonLisp",
	"Composer",
	"Concrete5",
	"Coq",
	"Dart",
	"Delphi",
	"DM",
	"Drupal",
	"Eagle",
	"Elisp",
	"Elixir",
	"EPiServer",
	"Erlang",
	"ExpressionEngine",
	"ExtJS-MVC",
	"Fancy",
	"Finale",
	"ForceDotCom",
	"Fortran",
	"FuelPHP",
	"gcov",
	"Go",
	"Gradle",
	"Grails",
	"GWT",
	"Haskell",
	"Idris",
	"Java",
	"Jboss",
	"Jekyll",
	"Joomla",
	"Jython",
	"Kohana",
	"LabVIEW",
	"Laravel4",
	"Leiningen",
	"LemonStand",
	"Lilypond",
	"Lithium",
	"Magento",
	"Maven",
	"Mercury",
	"MetaProgrammingSystem",
	"Meteor",
	"nanoc",
	"Node",
	"Objective-C",
	"OCaml",
	"Opa",
	"OpenCart",
	"OracleForms",
	"Packer",
	"Perl",
	"Phalcon",
	"PlayFramework",
	"Plone",
	"Prestashop",
	"Processing",
	"Python",
	"Qooxdoo",
	"Qt",
	"Rails",
	"R",
	"RhodesRhomobile",
	"ROS",
	"Ruby",
	"Rust",
	"Sass",
	"Scala",
	"SCons",
	"Scrivener",
	"Sdcc",
	"SeamGen",
	"SketchUp",
	"stella",
	"SugarCRM",
	"Swift",
	"Symfony2",
	"Symfony",
	"SymphonyCMS",
	"Target3001",
	"Tasm",
	"TeX",
	"Textpattern",
	"TurboGears2",
	"Typo3",
	"Umbraco",
	"Unity",
	"VisualStudio",
	"VVVV",
	"Waf",
	"WordPress",
	"Yeoman",
	"Yii",
	"ZendFramework",
	"Zephir"
]

TRUE = True
FALSE = False
BOOLEANS = [TRUE, FALSE]

NEWEST = "newest"
OLDEST = "oldest"
STARGAZERS = "stargazers"

SORT = [ NEWEST, OLDEST, STARGAZERS ]






import os.path, lib.json, lib.list, lib.ver, lib.git, collections

NPM       = 'package.json'
BOWER     = 'bower.json'
COMPONENT = 'component.json'

PM = [NPM, BOWER, COMPONENT]

VERSION_HASH   = 'version'


def unique(versions):
	if not versions : return None

	lib.check.VersionNotUniqueException(versions)

	version = next(iter(versions.values()))
	return version

def readpackagefiles():

	versions = {}
	for pm in PM:
		with lib.json.proxy(pm, 'r', default = {}) as conf:
			if VERSION_HASH in conf:
				old = conf[VERSION_HASH]
				lib.check.OldSemverVersionTagNotValidException(old, pm)
				versions[pm] = old

	return versions

def writeversion(version, files = PM):

	hook = collections.OrderedDict

	for pm in files:
		with lib.json.proxy(pm, 'w', object_pairs_hook = hook) as conf:
			conf[VERSION_HASH] = version


def getversion():
	versions = readpackagefiles()
	return unique(versions)

def setversion(version):

	olds = readpackagefiles()
	old = unique(olds)

	if lib.ver.isspecial(version) :
		lib.check.CannotInferSemverVersionNumberException(old, version)
		version = lib.ver.resolve(old, version)

	else :
		lib.check.SemverVersionTagNotValidException(version)
		lib.check.NewSemverVersionTagNotGreaterException(version, old)

	writeversion(version, files = olds.keys())

	return version


def upload(version, message = None):
	version = lib.ver.PREFIX + version
	if message is None : message = version

	lib.git.add('--all', '.')
	lib.git.commit('-am', message)
	lib.git.pull()
	lib.git.push()
	lib.git.tag('-a', version, '-m', message)
	lib.git.push('--tags')



import lib.sys, lib.config

try :
	import urllib.request as urllib2
except :
	import urllib2

def get(user = None, passwd = None):

	user, passwd = lib.config.prompt_cred('mail.google.com', 'gmail')

	auth_handler = urllib2.HTTPBasicAuthHandler()
	auth_handler.add_password(
		realm = 'New mail feed',
		uri = 'https://mail.google.com',
		user = user,
		passwd = passwd
	)

	opener = urllib2.build_opener(auth_handler)
	feed = opener.open('https://mail.google.com/mail/feed/atom')

	contents = feed.read().decode('utf-8')

	i = contents.index('<fullcount>') + 11
	j = contents.index('</fullcount>')

	unread = contents[i:j]

	print(unread)


def open():
	lib.sys.call(['google-chrome', 'https://mail.google.com']);



"""
	Doc for sak.test.do
"""

import time , itertools , lib.fn , lib.args

def action ( a , b = 10 ) :
	"""
		Doc for sak.test.do.action
	"""
	print( a , b )

@lib.args.convert( times = int , interval = int )
def throttled ( times , interval ) :

	@lib.fn.throttle( times , interval )
	def fn ( i ) :
		print ( i , time.time() )

	for i in itertools.count( 1 ) : fn( i )



import inspect, sak, lib.pacman, lib.check, functools, lib.source , lib.predicate

def ensurefmt ( fmt , n ) :
	fmt = list( fmt )
	if not fmt : fmt = [ print ]
	while len( fmt ) < n : fmt.append( fmt[-1] )
	return tuple( fmt )


def walk ( R , fmt , *path ) :

	fmtR, fmtM, fmtA = ensurefmt(fmt, 3)

	_ , module = lib.sak.resolve( R , path , ['sak'] )

	if module == R : print( fmtR( module ) )

	elif inspect.ismodule( module ) : print( fmtM( module ) )

	else : print( fmtA( module ) )


def info ( *path ) :

	fmtR = functools.partial( lib.pacman.format , pred = inspect.ismodule )
	fmtM = functools.partial( lib.pacman.format , pred = lib.predicate.disjunction( [ inspect.isfunction , inspect.ismodule ] ) )
	fmtA = lambda A : inspect.formatargspec( *inspect.getfullargspec( A ) )

	walk( sak , [ fmtR , fmtM , fmtA ] , *path )


def doc ( *path ) :
	"""
		Print the doc of the specified element (default = sak root module)
	"""

	walk( sak , [ inspect.getdoc ] , *path )


def source ( *path , linenos = False , filename = False ) :
	"""
		Print the source of the specified element (default = sak root module)
	"""

	fmt = functools.partial(lib.source.pretty, linenos = linenos, filename = filename)

	walk( sak , [ fmt ] , *path )



# requires https://github.com/bevry/cson

import os, tempfile, json, lib.sys


def load ( fp ) :

	return loads( fp.read() )


def loads ( string ) :

	with tempfile.NamedTemporaryFile( "w", delete = False, suffix = ".cson" ) as tmp :

		tmp.write( string )

	out, err, p = lib.sys.call( ["cson2json", tmp.name] )

	os.remove( tmp.name )

	return json.loads( out.decode() )


def dump ( obj, fp ) :

	fp.write( dumps( obj ) )


def dumps ( obj ) :

	with tempfile.NamedTemporaryFile( "w", delete = False, suffix = ".json" ) as tmp :

		json.dump( obj, tmp )

	out, err, p = lib.sys.call( ["json2cson", tmp.name] )

	os.remove( tmp.name )

	return out.decode()



import lib.orddir, lib.file

def read(cwd = '.'):
	d, w = lib.orddir.read(cwd)
	print(d, w)


def move(item, where, cwd = '.'):
	beg = int(item)
	end = beg + 1
	shift(where, beg, end, cwd)


def zfill(z, cwd = '.'):
	d, _ = lib.orddir.read(cwd)
	mv = []
	for k, l in d.items():
		_k = lib.orddir.format(k, z)
		mv += [(f, _k + v) for f, v in l]

	lib.file.move(mv, cwd)


def shift(where, beg = '0', end = None, cwd = '.'):

	where, beg, end = lib.orddir.drange(where, beg, end)
	d, w = lib.orddir.read(cwd)
	n = where - beg
	mv, y = lib.orddir.shift(d, n, beg, end)

	lib.file.move(mv, cwd)
	zfill(str(max(w, y)), cwd)




import lib.sys

IGNORE = '.gitignore'

def do(action, *args, **kwargs):
	return lib.sys.call(['git', action] + list(args), stddefault = None, **kwargs)

def clone(*args, **kwargs):
	return do('clone', *args, **kwargs)

def pull(*args, **kwargs):
	return do('pull', *args, **kwargs)

def push(*args, **kwargs):
	return do('push', *args, **kwargs)

def commit(*args, **kwargs):
	return do('commit', *args, **kwargs)

def remote(*args, **kwargs):
	return do('remote', *args, **kwargs)

def add(*args, **kwargs):
	return do('add', *args, **kwargs)

def status(*args, **kwargs):
	return do('status', *args, **kwargs)

def diff(*args, **kwargs):
	return do('diff', *args, **kwargs)

def tag(*args, **kwargs):
	return do('tag', *args, **kwargs)

def log(*args, **kwargs):
	return do('log', *args, **kwargs)

def checkout(*args, **kwargs):
	return do('checkout', *args, **kwargs)

def submodule(*args, **kwargs):
	return do('submodule', *args, **kwargs)




import sys , lib.json , lib.args

def format ( *args , **kwargs ) :

	if args : lib.json.pretty( args , sys.stdout )
	if kwargs : lib.json.pretty( kwargs , sys.stdout )

def args ( files = None ) :

	files = lib.args.listify( files )

	for f in files :

		with lib.json.proxy( f ) as data :

			for key , val in data.items( ) :

				val = [ lib.args.escape( x ) for x in lib.args.listify( val ) ]
				val = " ".join( val )

				print( "--%s %s" % ( key , val ) , end = " " )

	print( )





try :
	import http.client as httplib
except :
	import httplib

try :
	import urllib.parse as urlparse
except :
	import urlparse

def access(url):
	p = urlparse.urlparse(url)
	conn = httplib.HTTPConnection(p.netloc)
	conn.request('HEAD', p.path)
	resp = conn.getresponse()
	return resp.status < 400


def url(domain, path = None, username = None, secure = False):

	fmt = "http"
	args = []

	if secure : fmt += "s"

	fmt += "://%s"

	if username is not None :
		fmt += "@%s"
		args.append(username)

	args.append(domain)

	if path is not None:
		fmt += "/%s"
		args.append(path)

	return fmt % tuple(args)



import lib.args, lib.sys, fileinput, itertools, getpass, lib.file

# polyfill for generator zip function

if hasattr( itertools , "izip" ) :
	_zip = itertools.izip

else :
	_zip = zip


def directories ( callable = None , iterable = None ) :

	iterable = lib.args.listify( iterable )
	callable = lib.args.listify( callable )

	callable = list( itertools.chain( *map( lib.args.split , callable ) ) )

	if not iterable :
		iterable = ( s[:-1] for s in fileinput.input( [] ) )

	for item in iterable :

		lib.sys.call( callable , stddefault = None , cwd = item )


def imap ( callable = None , iterable = None ) :

	iterable = lib.args.listify( iterable )
	callable = lib.args.listify( callable )

	callable = list( itertools.chain( *map( lib.args.split , callable ) ) )

	if not iterable :
		iterable = ( s[:-1] for s in fileinput.input( [] ) )

	for item in iterable :

		lib.sys.call( [ arg.format( item ) for arg in callable ] , stddefault = None )


def starmap ( callable = None , iterable = None ) :

	iterable = lib.args.listify( iterable )
	callable = lib.args.listify( callable )

	callable = list( itertools.chain( *map( lib.args.split , callable ) ) )

	if not iterable :
		iterable = ( s[:-1] for s in fileinput.input( [] ) )

	for item in iterable :

		argv = lib.args.split( item )

		args , kwargs = lib.args.parse( argv , [] , {} )

		lib.sys.call( [ arg.format( *args , **kwargs ) for arg in callable ] , stddefault = None )


@lib.args.convert( n = int )
def repeat ( item , n = -1 ) :

	"""
		Repeat given string n times. If n is negative then repeat given string an infinite number of times.
	"""

	if n < 0 :
		args = [ None ]
	else :
		args = [ None , n ]

	for _ in itertools.repeat( *args ) :
		print( item )


@lib.args.convert( n = int )
def password ( n = -1 ) :

	item = getpass.getpass('Password to repeat : ')
	repeat( item , n )


def izip ( callables = None , sep = " " ) :

	callables = lib.args.listify( callables )

	callables = map( lib.args.split , callables )

	iterables = [ lib.file.lineiterator( lib.sys.popen( callable ).stdout ) for callable in callables ]

	for t in _zip( *iterables ) :

		print ( *t , sep = sep )



import lib.git, lib.args, lib.json, lib.time, collections

def commit ( message, duration, authors = None ) :

	authors = lib.args.listify( authors )

	hook = collections.OrderedDict

	with lib.json.proxy( "contributions.json", mode = "w", default = [], object_pairs_hook = hook ) as contribs :

		contribs.append( hook( [

			( "authors", authors ),
			( "message", message ),
			( "duration", duration ),
			( "timestamp", str( lib.time.nanoseconds( ) ) )

		] ) )

	lib.git.add( "contributions.json" )
	lib.git.commit( "-m", message )


def manhours ( begin = None , end = None ) :

	if begin is None :
		begin = 0
	else :
		begin = lib.time.parsedmy( begin )

	if end is None :
		end = lib.time.nanoseconds( )
	else :
		end = lib.time.parsedmy( end )

	total = 0

	with lib.json.proxy( "contributions.json", mode = "r", default = [] ) as contribs :

		for contrib in contribs :

			t = int( contrib["timestamp"] )

			if t >= begin and t <= end :

				total += int( contrib["duration"] ) * len( contrib["authors"] )

	print( total )


def log ( begin = None , end = None ) :

	if begin is None :
		begin = 0
	else :
		begin = lib.time.parsedmy( begin )

	if end is None :
		end = lib.time.nanoseconds( )
	else :
		end = lib.time.parsedmy( end )

	with lib.json.proxy( "contributions.json", mode = "r", default = [] ) as contribs :

		for contrib in contribs :

			t = int( contrib["timestamp"] )

			if t >= begin and t <= end :

				authors = contrib["authors"]
				duration = contrib["duration"]

				print( lib.time.pretty( t ) , "~" , " , ".join( authors ) , "(" , duration , ")" )
				print( " ->" , contrib["message"] )
				print( "===" )



import lib.cpm, lib.config, lib.git, lib.http, os.path


def update():
	pred = lambda x : os.path.exists(x)
	lib.cpm.depforeach(lib.git.pull, pred, cwd = lambda p : os.path.join(lib.cpm.DEP + p))


def install(package = None, mode = None):

	lib.cpm.structure()

	if package is None:
		pred = lambda x : not os.path.exists(x)
		lib.cpm.depforeach(lib.git.clone, pred, cwd = lib.cpm.DEP)

	else:

		url = lib.repo.resolve(package, lib.cpm.VENDORS)

		if url is None:
			print('unable to find repo')
			return

		lib.git.clone(url, cwd = lib.cpm.DEP)

		if mode in SAVE:
			which = SAVE[mode]
			parts = os.path.split(p.path)
			name, _ = os.path.splitext(parts)
			with lib.cpm.package('w') as p:
				p[which][name] = url


def init():
	if not lib.cpm.configOK():
		with lib.cpm.package('w') as p:
			p['name'] = ''
			p['dep'] = {}
			p['dep-dev'] = {}
			p['version'] = '0.0.0'
			p['keywords'] = {}
			p['run'] = {}
			p['bugs'] = {}
			p['bugs']['url'] = ''
			p['author'] = ''
			p['description'] = ''
			p['repo'] = {}
			p['repo']['url'] = ''
			p['repo']['type'] = ''
			p['home'] = ''
			p['license'] = ''

	lib.cpm.structure()




try : from setuptools import setup
except ImportError : from distutils.core import setup

setup (

	name = 'sak' , version = '0.4.0' ,

	description = 'Swiss Army Knife',
	long_description = 'sak is a module, submodule and function based tool' ,

	author = 'aureooms' ,
	author_email = 'aurelien.ooms@gmail.com' ,
	url = 'https://github.com/aureooms/sak' ,
	license = 'LICENSE' ,

	install_requires = [
		'lxml' ,
		'semantic_version'
	] ,

	packages = [
		'sak' ,
		'sak.test' ,
		'lib' ,
		'lib.nice'
	] ,

	scripts = [ '$' ]
)



import os, os.path

def install(source = ".", name = None, www = "/var/www"):
	source = os.path.abspath(source)
	if name is None : name = os.path.basename(source)
	link_name = os.path.join(www, name)
	os.symlink(source, link_name)

def uninstall(name = None, www = "/var/www"):
	if name is None : name = os.path.basename(".")
	link_name = os.path.join(www, name)
	os.remove(link_name)



import subprocess, lib.file, lib.error, lib.check, lib.npm, lib.git

def npm ( *args , cwd = None ) :
	return subprocess.call(['npm'] + list(args) , cwd = cwd )

def publish( cwd = None ):
	return npm('publish', cwd = cwd)

def unpublish(*args, cwd = None ):
	return npm('unpublish', *args, cwd = cwd)

def build( cwd = None ):
	return npm('run', 'build', cwd = cwd)

def doc( cwd = None ):
	return npm('run', 'doc', cwd = cwd)

def test( cwd = None ):
	return npm('test', cwd = cwd)

def install(*args, cwd = None ):
	return npm('install', *args, cwd = cwd)

def clean():
	lib.file.rm('node_modules', 'coverage', 'doc', recursive = True, force = True)

def upload(version, message = None):
	lib.npm.upload(version, message)

try:

	import semantic_version


	def release(version, message = None):
		try :
			doc()
			build()
			version = lib.npm.setversion(version)
			lib.npm.upload(version, message)
			publish()

		finally :
			lib.git.checkout( "master" )

	def setversion(version):
		version = lib.npm.setversion(version)
		print(version)

	def getversion():
		version = lib.npm.getversion()
		print(version)


except ImportError as cause:

	e = lib.error.ModuleMissingException(cause, "semantic_version")

	release = lambda version, message = None : lib.error.throw(e)
	setversion = lambda version : lib.error.throw(e)
	getversion = lambda : lib.error.throw(e)




import os , sak , lib.git , lib.sak

def git(action, *args):
	path = os.path.join(__file__, '..', '..')
	path = os.path.abspath(path)
	lib.git.do(action, *args, cwd = path)

def pull(*args):
	git('pull', *args)

def push(*args):
	git('push', *args)

def commit(*args):
	git('commit', *args)

def add(*args):
	git('add', *args)

def status(*args):
	git('status', *args)

def diff(*args):
	git('diff', *args)


def main ( inp ) :

	"""
		>>> from sak.sak import *
		>>> main( [ 'ts' , 'd' , 'a' , '2' ] )
		2 10

	"""

	hierarchy , action , inp = lib.sak.findaction( sak , inp , ['sak'] )
	args , kwargs = lib.sak.assignarguments( hierarchy , action , inp )
	action( *args , **kwargs )



import lib.config, lib.git, lib.error, lib.check
import lib.github, lib.input, lib.http, lib.args
import re

DOMAIN = lib.github.DOMAIN
CONFIG_KEY = lib.github.CONFIG_KEY
LICENSES = lib.github.LICENSES
GITIGNORES = lib.github.GITIGNORES
TRUE = lib.github.TRUE
FALSE = lib.github.FALSE
BOOLEANS = lib.github.BOOLEANS
YOU = lib.github.YOU
NEWEST = lib.github.NEWEST
OLDEST = lib.github.OLDEST
STARGAZERS = lib.github.STARGAZERS
SORT = lib.github.SORT



def clone( repo, dest = None, username = None ):

	url = lib.http.url( DOMAIN, repo, username, secure = True )

	if dest is not None :
		return lib.git.clone( url, dest )
	else :
		return lib.git.clone( url )


def new(name, org = None, team_id = None, username = None, password = None, auto_init = FALSE, private = FALSE, description = None, homepage = None, has_issues = TRUE, has_wiki = TRUE, has_downloads = TRUE, gitignore_template = None, license_template = None):

	lib.check.OptionNotInListException( "private", private, BOOLEANS )
	lib.check.OptionNotInListException( "has_issues", has_issues, BOOLEANS )
	lib.check.OptionNotInListException( "has_wiki", has_wiki, BOOLEANS )
	lib.check.OptionNotInListException( "has_downloads", has_downloads, BOOLEANS )
	lib.check.OptionNotInListException( "auto_init", auto_init, BOOLEANS )
	lib.check.OptionNotInListException( "gitignore_template", gitignore_template, GITIGNORES )
	lib.check.OptionNotInListException( "license_template", license_template, LICENSES )

	username, password = lib.github.credentials(username, password)


	parameters = {
		"name" : name,
		"description" : description,
		"homepage" : homepage,
		"private": private,
		"has_issues" : has_issues,
		"has_wiki" : has_wiki,
		"has_downloads" : has_downloads,
		"team_id": team_id,
		"auto_init" : auto_init,
		"gitignore_template" : gitignore_template,
		"license_template" : license_template
	}


	if org is None :
		url = ( "user" , "repos" )
	else :
		url = ( "orgs" , org , "repos" )

	_, _, p = lib.github.post( url , data = parameters , username = username , password = password , stddefault = None )
	print()
	lib.check.SubprocessReturnedFalsyValueException(p.args, p.returncode)


def group ( *names ) :

	username, password = lib.github.credentials( None, None )

	org = None
	team_id = None
	auto_init = FALSE
	private = FALSE
	description = None
	homepage = None
	has_issues = TRUE
	has_wiki = TRUE
	has_downloads = TRUE
	gitignore_template = None
	license_template = None


	for name in names :
		new(name, org, team_id, username, password, auto_init, private, description, homepage, has_issues, has_wiki, has_downloads, gitignore_template, license_template)


def list ( target = YOU , name = None , t = None , format = "{full_name}" , username = None , password = None ) :

	for repo in lib.github.list( target , name , t , username , password ) :
		print( format.format( **repo ) )


def download ( target = YOU, name = None, t = None, username = None, password = None, prompt = True, prefix = "", suffix = "", regexp = "" ) :

	for repo in lib.github.list( target, name, t, username, password):

		repo = repo["full_name"]

		take = True

		take = take and ( not prefix or repo.startswith( prefix ) )
		take = take and ( not suffix or repo.endswith( suffix ) )
		take = take and ( not regexp or re.match( regexp, repo ) is not None )

		if take and ( not prompt or lib.input.yesorno( "clone '%s'?" % repo ) ) :
			clone( repo, username = username )


def delete(owner, repo, username = None, password = None):
	username, password = lib.github.credentials(username, password)

	url = ( "repos" , owner , repo )
	_, _, p = lib.github.delete( url, username = username, password = password, stddefault = None)
	print()
	lib.check.SubprocessReturnedFalsyValueException(p.args, p.returncode)



def issues ( owner = None , repo = None , number = None , user = False , org = None , username = None , password = None , milestone = None , filter = None , state = None , creator = None , assignee = None , mentioned = None , labels = None , sort = None , direction = None , since = None ) :

	for issue in lib.args.forward( lib.github.issues , locals( ) ) :

		print( issue )


def createissue ( owner, repo, title, body = None, assignee = None, milestone = None, labels = None, username = None, password = None ) :

	print( lib.args.forward( lib.github.createissue , locals( ) ) )


def editissue ( owner , repo , number , title = None , body = None , assignee = None , state = None , milestone = None , labels = None , username = None , password = None ) :

	print( lib.args.forward( lib.github.editissue , locals( ) ) )


def closeissues ( owner , repo , *issuenos , username = None , password = None ) :

	for response in lib.args.forward( lib.github.closeissues , locals( ) ) :

		print( response )


def comments ( owner , repo , id = None , number = None , sort = None , direction = None , since = None , username = None , password = None ) :

	print( lib.args.forward( lib.github.comments , locals( ) ) )


def createcomment ( owner , repo , number , body , username = None , password = None ) :

	print( lib.args.forward( lib.github.createcomment , locals( ) ) )


def editcomment ( owner , repo , id , body , username = None , password = None ) :

	print( lib.args.forward( lib.github.editcomment , locals( ) ) )


def deletecomment ( owner , repo , id , username = None , password = None ) :

	print( lib.args.forward( lib.github.deletecomment , locals( ) ) )


def labels ( owner, repo, name = None, issue = None, username = None, password = None ) :

	print( lib.args.forward( lib.github.labels , locals( ) ) )


def createlabel ( owner, repo, name, color, username = None, password = None ) :

	print( lib.args.forward( lib.github.createlabel , locals( ) ) )


def updatelabel ( owner, repo, oldname, newname, color, username = None, password = None ) :

	print( lib.args.forward( lib.github.updatelabel , locals( ) ) )


def deletelabel ( owner, repo, name, username = None, password = None ) :

	print( lib.args.forward( lib.github.deletelabel , locals( ) ) )


def addlabels ( owner, repo, issue, labels = None, username = None, password = None ) :

	print( lib.args.forward( lib.github.addlabels , locals( ) ) )


def removelabel ( owner, repo, issue, label, username = None, password = None ) :

	print( lib.args.forward( lib.github.removelabel , locals( ) ) )


def updatelabels ( owner, repo, issue, labels = None, username = None, password = None ) :

	print( lib.args.forward( lib.github.updatelabels , locals( ) ) )


def removealllabels ( owner, repo, issue, username = None, password = None ) :

	print( lib.args.forward( lib.github.removealllabels , locals( ) ) )


def milestonelabels ( owner, repo, milestone, username = None, password = None ) :

	print( lib.args.forward( lib.github.milestonelabels , locals( ) ) )


def milestones ( owner , repo , number = None , state = None , sort = None , direction = None , username = None , password = None ) :

	print( lib.args.forward( lib.github.milestones , locals( ) ) )


def createmilestone ( owner , repo , title , state = None , description = None , due_on = None , username = None , password = None ) :

	print( lib.args.forward( lib.github.createmilestone , locals( ) ) )


def updatemilestone ( owner , repo , number , title , state = None , description = None , due_on = None , username = None , password = None ) :

	print( lib.args.forward( lib.github.updatemilestone , locals( ) ) )


def deletemilestone ( owner , repo , number , username = None , password = None ) :

	print( lib.args.forward( lib.github.deletemilestone , locals( ) ) )


def listforks ( owner, repo, sort = NEWEST, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	if username is not None :
		username, password = lib.github.credentials( username, password )

	url = ( "repos", owner, repo, "forks" )

	parameters = dict( sort = sort )

	_, _, p = lib.github.get( url, data = parameters, username = username, password = password, stddefault = None )
	print()
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )


def fork ( owner, repo, organization = None, username = None, password = None ) :

	"""
		https://developer.github.com/v3/issues/labels/
	"""

	username, password = lib.github.credentials( username, password )

	url = ( "repos", owner, repo, "forks" )

	parameters = dict( organization = organization )

	_, _, p = lib.github.post( url, data = parameters, username = username, password = password, stddefault = None )
	print()
	lib.check.SubprocessReturnedFalsyValueException( p.args, p.returncode )


def patch ( owner, repo, name, username = None, password = None, private = FALSE, description = None, homepage = None, has_issues = TRUE, has_wiki = TRUE, has_downloads = TRUE, default_branch = None ) :

	lib.check.OptionNotInListException( "private", private, BOOLEANS )
	lib.check.OptionNotInListException( "has_issues", has_issues, BOOLEANS )
	lib.check.OptionNotInListException( "has_wiki", has_wiki, BOOLEANS )
	lib.check.OptionNotInListException( "has_downloads", has_downloads, BOOLEANS )

	username, password = lib.github.credentials(username, password)


	parameters = {
		"name" : name,
		"description" : description,
		"homepage" : homepage,
		"private": private,
		"has_issues" : has_issues,
		"has_wiki" : has_wiki,
		"has_downloads" : has_downloads,
		"default_branch": default_branch
	}


	url = ( "repos", owner, repo )

	_, _, p = lib.github.patch( url , data = parameters, username = username, password = password, stddefault = None)
	print()
	lib.check.SubprocessReturnedFalsyValueException(p.args, p.returncode)



def archive ( username , forks = False , gist = True , metadata = True ) :
	"""
		All credits go to Filippo Valsorda (https://filippo.io).

		original -> https://filippo.io/archive-your-github-repo-and-data/

		# This is free and unencumbered software released into the public domain.

		# Anyone is free to copy, modify, publish, use, compile, sell, or
		# distribute this software, either in source code form or as a compiled
		# binary, for any purpose, commercial or non-commercial, and by any
		# means.

		# In jurisdictions that recognize copyright laws, the author or authors
		# of this software dedicate any and all copyright interest in the
		# software to the public domain. We make this dedication for the benefit
		# of the public at large and to the detriment of our heirs and
		# successors. We intend this dedication to be an overt act of
		# relinquishment in perpetuity of all present and future rights to this
		# software under copyright law.

		# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
		# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
		# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
		# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
		# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
		# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
		# OTHER DEALINGS IN THE SOFTWARE.

		# For more information, please refer to <http://unlicense.org/>

		usage: gh_dump.py [-h] [--forks] [--no-gist] [--no-metadata] username

		Dump an user's public GitHub data into current directory.

		positional arguments:
		username the GH username

		optional arguments:
		-h, --help show this help message and exit
		--forks git clone also forks (default is don't)
		--no-gist don't download user gists (default is do)
		--no-metadata don't download user metadata (default is do)
	"""

	from urllib.request import urlopen
	from subprocess import call
	import json
	import re
	import os.path

	def clear_url(url):
		return re.sub(r'\{[^\}]*\}', '', url)

	data = urlopen('https://api.github.com/users/' + args.user).read()
	user = json.loads(data.decode('utf-8'))
	if args.metadata:
		with open('user.json', 'wb') as f:
			f.write(data)

	data = urlopen(clear_url(user['repos_url'])).read()
	repos = json.loads(data.decode('utf-8'))
	if args.metadata:
		with open('repos.json', 'wb') as f:
			f.write(data)
	for repo in repos:
		if not repo['fork']:
			call(['git', 'clone', repo['clone_url']])
		elif args.forks:
			if not os.path.exists('forks'):
				os.makedirs('forks')
			call(['git', 'clone', repo['clone_url'], os.path.join('forks', repo['name'])])

	data = urlopen(clear_url(user['gists_url'])).read()
	gists = json.loads(data.decode('utf-8'))
	if args.metadata:
		with open('gists.json', 'wb') as f:
			f.write(data)
	if args.gists:
		if not os.path.exists('gists'):
			os.makedirs('gists')
		for gist in gists:
			call(['git', 'clone', gist['git_pull_url'], os.path.join('gists', gist['id'])])

	if args.metadata:
		for name in ['received_events', 'events', 'organizations', 'followers', 'starred', 'following', 'subscriptions']:
			data = urlopen(clear_url(user[name + '_url'])).read()
			with open(name + '.json', 'wb') as f:
				f.write(data)



def migrateissues ( owner , origin , destination , *issuenos , username = None , password = None ) :

	username , password = lib.github.credentials( username , password )

	print( "fetch issues to migrate" )

	issuestomigrate = []

	for number in issuenos :

		out = lib.github.issues( owner , origin , number , username = username , password = password )

		lib.github.validate( out )

		issuestomigrate.append( out )


	print( "compute labels and milestones to migrate" )

	labelstomigrate = {}

	milestonestomigrate = {}

	for issue in issuestomigrate :

		if "labels" in issue and issue["labels"] is not None :

			for label in issue["labels"] :

				labelstomigrate.setdefault( label["name"] , label )

		if "milestone" in issue and issue["milestone"] is not None :

			milestone = issue["milestone"]

			milestonestomigrate.setdefault( milestone["title"] , milestone )


	print( "remove already existing labels from the migrate list" )

	labelsalreadythere = lib.github.labels( owner , destination , username = username , password = password )

	lib.github.validate( labelsalreadythere )

	for label in labelsalreadythere :

		labelstomigrate.pop( label["name"] , None )


	print( "remove already existing milestones from the migrate list and save them to map" )

	milestonesmap = {}

	milestonesalreadythere = lib.github.milestones( owner , destination , username = username , password = password )

	lib.github.validate( milestonesalreadythere )

	for milestone in milestonesalreadythere :

		title = milestone["title"]

		milestonefrom = milestonestomigrate.get( title , None )

		if milestonefrom is not None :

			milestonesmap[milestonefrom["number"]] = milestone["number"]

			milestonestomigrate.pop( title )


	print( "migrate labels" )

	for name , label in labelstomigrate.items( ) :

		color = label["color"]

		lib.github.validate( lib.github.createlabel( owner , destination , name , color , username = username , password = password ) )


	print( "migrate milestones and save map between old numbers and new ones" )

	for title , milestone in milestonestomigrate.items( ) :

		parameters = lib.dict.select( milestone , [ "state" , "description" , "due_on" ] )

		number = milestone["number"]

		out = lib.github.createmilestone( owner , destination , title , username = username , password = password , **parameters )

		lib.github.validate( out )

		milestonesmap[number] = out["number"]


	print( "migrate issues and save map between old numbers and new ones" )

	issuemap = {}

	for issue in issuestomigrate :

		parameters = {}

		if "milestone" in issue and issue["milestone"] is not None :

			number = issue["milestone"]["number"]

			parameters["milestone"] = milestonesmap[number]

		if "labels" in issue and issue["labels"] is not None :

			parameters["labels"] = [ label["name"] for label in issue["labels"] ]

		if "assignee" in issue and issue["assignee"] is not None :

			parameters["assignee"] = issue["assignee"]["login"]

		if "body" in issue and issue["body"] is not None :

			parameters["body"] = issue["body"]

		title = issue["title"]

		out = lib.github.createissue( owner , destination , title , username = username , password = password , **parameters )

		lib.github.validate( out )

		issuemap[issue["number"]] = out


	print( "fetch comments if necessary" )

	commentstomigrate = {}

	for issue in issuestomigrate :

		if "comments" in issue and issue["comments"] > 0 :

			number = issue["number"]

			commentstomigrate[number] = lib.github.comments( owner , origin , number = number , username = username , password = password )

			lib.github.validate( commentstomigrate[number] )


	print( "migrate comments" )

	for number , listofcomments in commentstomigrate.items( ) :

		number = issuemap[number]["number"]

		for comment in listofcomments :

			body = comment["body"]

			lib.github.validate( lib.github.createcomment( owner , destination , number , body , username = username , password = password ) )


	print( "add comment to say that it was migrated to destination" )
	print( "add comment to say that it was migrated from origin" )

	for issuefrom in issuestomigrate :

		numberfrom = issuefrom["number"]

		issueto = issuemap[numberfrom]

		numberto = issueto["number"]

		bodyfrom = "migrated to %s" % issueto["html_url"]

		bodyto = "migrated from %s" % issuefrom["html_url"]

		lib.github.validate( lib.github.createcomment( owner , origin , numberfrom , bodyfrom , username = username , password = password ) )

		lib.github.validate( lib.github.createcomment( owner , destination , numberto , bodyto , username = username , password = password ) )


	print( "close issues from origin" )

	for issue in issuestomigrate :

		parameters = {}

		parameters["title"] = issue.get( "title" , None )

		parameters["body"] = issue.get( "body" , None )

		parameters["state"] = "closed"

		if "assignee" in issue and issue["assignee"] is not None :
			parameters["assignee"] = issue["assignee"]["login"]

		if "milestone" in issue and issue["milestone"] is not None :
			parameters["milestone"] = issue["milestone"]["number"]

		if "labels" in issue and issue["labels"] is not None :
			parameters["labels"] = [ label["name"] for label in issue["labels"] ]

		lib.github.validate( lib.github.editissue( owner , origin , issue["number"] , username = username , password = password , **parameters ) )




import getpass, ftplib

class FTP(ftplib.FTP):

	def __init__(self, *args):
		ftplib.FTP.__init__(self, *args)
		self.FILE = '-'
		self.DIR = 'd'
		self.LINK = 'l'

	def storbinary(self, path, fd):
		super(FTP, self).storbinary('STOR %s' % path, fd)

	def chmod(self, mod, path):
		return self.sendcmd('SITE CHMOD %s %s' % (mod, path))

	def poke(self, path):
		if len(path) == 0 or path[0] != '/' : path = self.pwd() + '/' + path
		return (path, self.nlst(path))

	def isfile(self, path):
		path, l = self.poke(path)
		return len(l) == 1 and l[0] == path

	def isdir(self, path):
		path, l = self.poke(path)
		return len(l) > 0 and l[0] != path

	def ls(self, path = '.'):
		if path == '..' : path = '../'
		d = []
		self.retrlines('LIST %s' % path, lambda x : d.append((x[0], ''.join(x.split()[8:]))))
		return d

	def loginprompt(self, config):
		print(self.connect(config['host']))
		username = config['username'] if 'username' in config else input('Username for \'ftp://%s\' : ' % (config['host']))
		password = getpass.getpass('Password for \'ftp://%s@%s\' : ' % (username, config['host']))
		print(self.login(username, password))
		print(self.cwd(config['root']))
